from typing import Union
from pathlib import Path
import minio
import os
from datetime import timedelta
from .errors import FileUploadError, InvalidCredentialsError, BucketDoesNotExistsError
from zipfile import ZipFile
from .utils import api_error, Progress, get_all_file_paths, validate_base_url, validate_platform_id, \
    validate_namespace_id, validate_deployment_name, validate_username
import requests
import json

STORAGE_URL = "datalake.zbl-aws.zeblok.com:9000"
CROSS_CLOUD_SERVICE_URL = "http://cross-cloud-services-673735588.us-east-2.elb.amazonaws.com/upload"


class DeployModel:
    __slots__ = ['__CROSS_CLOUD_SERVICE_URL', '__storage_url', '__bucket_name', '__base_url', '__token', '__username']

    def __init__(self, base_url: str, token: str, bucket_name: str, username: str, storage_url: str):
        validate_base_url(base_url=base_url)
        if type(token) is not str:
            raise InvalidCredentialsError('token can only be of type String')
        elif token == '':
            raise InvalidCredentialsError('token cannot empty')
        elif token.split()[0] != 'Bearer':
            raise InvalidCredentialsError('Please pass a valid Bearer token')

        validate_username(username=username)

        self.__bucket_name = bucket_name
        self.__username = username
        self.__storage_url = storage_url
        self.__CROSS_CLOUD_SERVICE_URL = CROSS_CLOUD_SERVICE_URL
        self.__base_url = 'https://' + base_url
        self.__token = token

    @property
    def token(self):
        return self.__token

    @property
    def base_url(self):
        return self.__base_url

    @property
    def cross_cloud_service_url(self):
        return self.__CROSS_CLOUD_SERVICE_URL

    def get_datacenters(self) -> [dict]:
        response = requests.get(
            f"{self.__base_url}/api/v1/dataCenter/public/", data=json.dumps({"isActive": True}),
            headers={'Authorization': self.__token, 'Content-Type': 'application/json', 'request-from': 'Ai-MicroCloud'}
        )
        if response.status_code == 200 and response.json()['success']:
            datacenters = []
            for data in response.json()['data']:
                datacenters.append({'_id': data['_id'], 'name': data['name']})
            return datacenters
        else:
            api_error(status_code=response.status_code, message=response.text)

    def get_namespaces(self) -> [dict]:
        response = requests.get(
            f"{self.__base_url}/api/v1/namespaces/org/", data=json.dumps({"state": "ready"}),
            headers={'Authorization': self.__token, 'Content-Type': 'application/json', 'request-from': 'Ai-MicroCloud'}
        )
        if response.status_code == 200 and response.json()['success']:
            return response.json()['data']
        else:
            api_error(status_code=response.status_code, message=response.text)

    @staticmethod
    def prepare_model_zip(model_folder_path: Path) -> Path:
        print("\nPreparing model zip")
        model_zipfile_path = model_folder_path.parent.joinpath(f'{model_folder_path.name.lower()}.zip')
        with ZipFile(model_zipfile_path, 'w') as zip:
            file_paths = get_all_file_paths(directory=model_folder_path)
            for file in file_paths:
                zip.write(filename=file.as_posix(), arcname=file.relative_to(model_folder_path))
        print("Model zip prepared")
        return model_zipfile_path

    def upload_file(self, file_name: Path, secret_key: str) -> str:
        try:
            client = minio.Minio(
                endpoint=self.__storage_url, access_key=self.__username, secret_key=secret_key, secure=False
            )

            if not client.bucket_exists(self.__bucket_name):
                raise BucketDoesNotExistsError(f'{self.__bucket_name} does not exists')

            client.fput_object(
                bucket_name=self.__bucket_name, object_name=f"{file_name.name}",
                file_path=file_name.as_posix(), content_type="application/zip",
                progress=Progress(), part_size=5 * 1024 * 1024
            )

            url = client.presigned_get_object(
                bucket_name=self.__bucket_name, object_name=f"{file_name.name}",
                expires=timedelta(hours=3)
            )
        finally:
            os.remove(file_name)
        return url

    def call_cross_cloud_service(
            self, cross_cloud_service_url: str, presigned_get_url: str, image_name: str, file_name: str,
            image_id: Union[str, None], deployment_type: str, autodeploy: bool = False,
            namespace_id: Union[None, str] = None, platform_id: Union[None, str] = None,
            deployment_name: Union[None, str] = None
    ):
        if autodeploy:
            validate_platform_id(platform_id=platform_id)
            validate_namespace_id(namespace_id=namespace_id)
            validate_deployment_name(deployment_name=deployment_name)

        response = requests.post(
            cross_cloud_service_url,
            headers={'Authorization': self.__token, 'Content-Type': 'application/json',
                     'request-from': 'Ai-MicroCloud'},
            data=json.dumps({
                "imageName": image_name, "url": presigned_get_url, "filename": file_name,
                "portalUrl": f"{self.__base_url}", "imageId": image_id,
                "autoDeploy": autodeploy, "namespaceId": None if not autodeploy else namespace_id,
                "platform": None if not autodeploy else platform_id,
                "deploymentName": None if not autodeploy else deployment_name,
                "deploymentType": deployment_type
            })
        )
        if response.status_code == 200:
            if response.json()['success']:
                return
            else:
                raise FileUploadError(response.json()['job'])
        else:
            api_error(status_code=response.status_code, message=response.text)



from .errors import InvalidCredentialsError, AuthenticationError, ServerError
from .utils import validate_base_url, validate_username


class Auth:
    __slots__ = ['__username', '__password', 'base_url', '__token']

    def __init__(self, username: str, password: str, base_url: str):
        validate_username(username=username)
        self.__username = username

        if type(password) is not str:
            raise InvalidCredentialsError('password can only be of type String')
        if password == '':
            raise InvalidCredentialsError('password cannot empty')
        self.__password = password

        validate_base_url(base_url=base_url)
        self.base_url = 'https://' + base_url

        self.__token = None

    def __fetch_token(self):
        from requests import post
        from json import dumps

        response = post(
            f"{self.base_url}/authservice/api/v1/auth/cli",
            headers={'Content-Type': 'application/json'},
            data=dumps({'email': self.__username, 'password': self.__password})
        )
        if response.status_code == 200:
            self.__token = "Bearer " + response.json()['data']
        elif response.status_code == 401:
            raise AuthenticationError(response.json()['message'])
        else:
            raise ServerError(f"Status code = {response.status_code}")

    def get_username(self) -> str:
        return self.__username

    def get_password(self) -> str:
        return self.__password

    def get_token(self) -> str:
        if self.__token is None:
            self.__fetch_token()
        return self.__token



from typing import Union
from .deploy import DeployModel
from pathlib import Path
import os
import time
from .errors import InvalidModelFolder, FileUploadError, ModelDeploymentError
from .utils import api_error, get_all_file_paths, validate_model_pipeline, validate_username, validate_platform_id, \
    validate_namespace_id, validate_deployment_name, validate_model_folder
import requests
import json


class ModelAPI(DeployModel):
    def __init__(self, base_url: str, token: str, bucket_name: str, username: str, storage_url: str):
        super(ModelAPI, self).__init__(
            base_url=base_url, token=token, bucket_name=bucket_name, username=username, storage_url=storage_url
        )

    def get_model_pipelines(self) -> [dict]:
        response = requests.get(
            f"{self.base_url}/api/v1/aimodel/", data=json.dumps({"state": "ready"}),
            headers={'Authorization': self.token, 'Content-Type': 'application/json'}
        )
        if response.status_code == 200 and response.json()['success']:
            image_names = []
            for data in response.json()['data']:
                image_names.append({'_id': data['_id'], 'imageName': data['imageName']})
            return image_names
        else:
            api_error(status_code=response.status_code, message=response.text)

    @staticmethod
    def __validate_folder_format(model_folder_path: Path) -> None:
        validate_model_folder(model_folder_path=model_folder_path)
        for filepath in get_all_file_paths(directory=model_folder_path):
            if filepath.name == 'Dockerfile' and filepath.parent.name == model_folder_path.name:
                return
        raise InvalidModelFolder(f"Invalid BentoML folder format: {model_folder_path}")

    def __register_image_name(self, image_name: str):
        response = requests.post(
            f"{self.base_url}/api/v1/aimodel/", data=json.dumps({"imageName": image_name, "type": "bentoml"}),
            headers={'Authorization': self.token, 'Content-Type': 'application/json'}
        )
        print(response.request.body)
        if response.status_code == 201:
            if response.json()['success']:
                return response.json()['data']['_id']
            else:
                raise FileUploadError(response.json()['message'])
        else:
            api_error(status_code=response.status_code, message=response.text)

    def upload_model(
            self, model_folder_path: str, username: str, storage_secret_key: str,
            autodeploy: bool = False, namespace_id: Union[None, str] = None, platform_id: Union[None, str] = None,
            deployment_name: Union[None, str] = None,
    ) -> str:
        validate_username(username=username)
        folder_path = Path(model_folder_path)
        self.__validate_folder_format(model_folder_path=folder_path)
        model_zipfile_path = self.prepare_model_zip(model_folder_path=folder_path)
        presigned_url = self.upload_file(file_name=model_zipfile_path, secret_key=storage_secret_key)
        image_name = f"zeblok/{model_zipfile_path.stem}:{int(time.time())}".lower()
        image_id = self.__register_image_name(image_name=image_name)
        self.call_cross_cloud_service(
            cross_cloud_service_url=self.cross_cloud_service_url, presigned_get_url=presigned_url,
            image_name=image_name, file_name=model_zipfile_path.name, image_id=image_id, autodeploy=autodeploy,
            namespace_id=namespace_id, deployment_name=deployment_name, platform_id=platform_id,
            deployment_type='aimodel'
        )
        print(
            f"\nSuccessfully uploaded the Model folder | Filename: {model_zipfile_path.name}, Image Name: {image_name}"
        )
        return image_name

    def deploy_model(self, deployment_name: str, namespace_id: str, platform_id: str, model_pipeline: str):
        validate_deployment_name(deployment_name=deployment_name)
        validate_namespace_id(namespace_id=namespace_id)
        validate_platform_id(platform_id=platform_id)
        validate_model_pipeline(model_pipeline=model_pipeline)

        if model_pipeline not in [img_nm['imageName'] for img_nm in self.get_model_pipelines()]:
            raise ValueError(f'Image Name: {model_pipeline} not found in the database')

        response = requests.post(
            f"{self.base_url}/api/v1/k8deployments/",
            headers={'Authorization': self.token, 'Content-Type': 'application/json'},
            data=json.dumps({
                "imageName": model_pipeline, "nodePreference": "NO PREFERENCE", "kioskId": None,
                "platform": platform_id,
                "namespaceId": namespace_id, "deploymentName": deployment_name
            })
        )

        if response.status_code == 201:
            if response.json()['success']:
                print(
                    f"\n{response.json()['message']}"
                )
                return response.json()['success']
            else:
                raise ModelDeploymentError(response.json()['message'])
        else:
            api_error(status_code=response.status_code, message=response.text)



"""
This module implements a progress printer while communicating with MinIO server
:copyright: (c) 2018 by MinIO, Inc.
:license: Apache 2.0, see LICENSE for more details.
"""

import sys
import time
from queue import Empty, Queue
from threading import Thread
from pathlib import Path
from os import walk
import os
from .errors import AuthenticationError, ServerError, InvalidCredentialsError, InvalidBaseURL, InvalidStorageURL

_BAR_SIZE = 20
_KILOBYTE = 1024
_FINISHED_BAR = '#'
_REMAINING_BAR = '-'

_UNKNOWN_SIZE = '?'
_STR_MEGABYTE = ' MB'

_HOURS_OF_ELAPSED = '%d:%02d:%02d'
_MINUTES_OF_ELAPSED = '%02d:%02d'

_RATE_FORMAT = '%5.2f'
_PERCENTAGE_FORMAT = '%3d%%'
_HUMANINZED_FORMAT = '%0.2f'

_DISPLAY_FORMAT = '|%s| %s/%s %s [elapsed: %s left: %s, %s MB/sec]'

_REFRESH_CHAR = '\r'


def get_all_file_paths(directory):
    file_paths = []
    for root, directories, files in walk(directory):
        for filename in files:
            file_paths.append(Path(root).joinpath(filename))
    return file_paths


def seconds_to_time(seconds):
    """
    Consistent time format to be displayed on the elapsed time in screen.
    :param seconds: seconds
    """
    minutes, seconds = divmod(int(seconds), 60)
    hours, m = divmod(minutes, 60)
    if hours:
        return _HOURS_OF_ELAPSED % (hours, m, seconds)
    else:
        return _MINUTES_OF_ELAPSED % (m, seconds)


def format_string(current_size, total_length, elapsed_time):
    """
    Consistent format to be displayed on the screen.
    :param current_size: Number of finished object size
    :param total_length: Total object size
    :param elapsed_time: number of seconds passed since start
    """

    n_to_mb = current_size / _KILOBYTE / _KILOBYTE
    elapsed_str = seconds_to_time(elapsed_time)

    rate = _RATE_FORMAT % (
            n_to_mb / elapsed_time) if elapsed_time else _UNKNOWN_SIZE
    frac = float(current_size) / total_length
    bar_length = int(frac * _BAR_SIZE)
    bar = (_FINISHED_BAR * bar_length +
           _REMAINING_BAR * (_BAR_SIZE - bar_length))
    percentage = _PERCENTAGE_FORMAT % (frac * 100)
    left_str = (
        seconds_to_time(
            elapsed_time / current_size * (total_length - current_size))
        if current_size else _UNKNOWN_SIZE)

    humanized_total = _HUMANINZED_FORMAT % (
            total_length / _KILOBYTE / _KILOBYTE) + _STR_MEGABYTE
    humanized_n = _HUMANINZED_FORMAT % n_to_mb + _STR_MEGABYTE

    return _DISPLAY_FORMAT % (bar, humanized_n, humanized_total, percentage,
                              elapsed_str, left_str, rate)


def api_error(status_code: int, message: str):
    if status_code == 401:
        raise AuthenticationError("User not authenticated. Please check your __api_access_key and api_access_secret")
    else:
        raise ServerError(f"Status code = {status_code}\n{message}")


class Progress(Thread):
    """
        Constructs a :class:`Progress` object.
        :param interval: Sets the time interval to be displayed on the screen.
        :param stdout: Sets the standard output
        :return: :class:`Progress` object
    """

    def __init__(self, interval=1, stdout=sys.stdout):
        Thread.__init__(self)
        self.daemon = True
        self.total_length = 0
        self.interval = interval
        self.object_name = None

        self.last_printed_len = 0
        self.current_size = 0

        self.display_queue = Queue()
        self.initial_time = time.time()
        self.stdout = stdout
        self.start()
        self.prefix = None

    def set_meta(self, total_length, object_name):
        """
        Metadata settings for the object. This method called before uploading
        object
        :param total_length: Total length of object.
        :param object_name: Object name to be showed.
        """
        self.total_length = total_length
        self.object_name = object_name
        self.prefix = self.object_name + ': ' if self.object_name else ''

    def run(self):
        displayed_time = 0
        while True:
            try:
                # display every interval secs
                task = self.display_queue.get(timeout=self.interval)
            except Empty:
                elapsed_time = time.time() - self.initial_time
                if elapsed_time > displayed_time:
                    displayed_time = elapsed_time
                self.print_status(current_size=self.current_size,
                                  total_length=self.total_length,
                                  displayed_time=displayed_time,
                                  prefix=self.prefix)
                continue

            current_size, total_length = task
            displayed_time = time.time() - self.initial_time
            self.print_status(current_size=current_size,
                              total_length=total_length,
                              displayed_time=displayed_time,
                              prefix=self.prefix)
            self.display_queue.task_done()
            if current_size == total_length:
                # once we have done uploading everything return
                self.done_progress()
                return

    def update(self, size):
        """
        Update object size to be showed. This method called while uploading
        :param size: Object size to be showed. The object size should be in
                     bytes.
        """
        if not isinstance(size, int):
            raise ValueError('{} type can not be displayed. '
                             'Please change it to Int.'.format(type(size)))

        self.current_size += size
        self.display_queue.put((self.current_size, self.total_length))

    def done_progress(self):
        self.total_length = 0
        self.object_name = None
        self.last_printed_len = 0
        self.current_size = 0

    def print_status(self, current_size, total_length, displayed_time, prefix):
        formatted_str = prefix + format_string(
            current_size, total_length, displayed_time)
        self.stdout.write(_REFRESH_CHAR + formatted_str + ' ' *
                          max(self.last_printed_len - len(formatted_str), 0))
        self.stdout.flush()
        self.last_printed_len = len(formatted_str)


def validate_secret_key(key_name: str, secret_key: str):
    if type(secret_key) is not str:
        raise InvalidCredentialsError(f'{key_name} can only be of type String')
    if secret_key == '':
        raise InvalidCredentialsError(f'{key_name} cannot empty')


def validate_username(username: str):
    if type(username) is not str:
        raise InvalidCredentialsError('username can only be of type String')
    if username == '':
        raise InvalidCredentialsError('username cannot empty')


def validate_url(url_name: str, url: str):
    if type(url) is not str:
        raise InvalidBaseURL(f'{url_name} can only be of type String')
    if url == '':
        raise InvalidBaseURL(f'{url_name} cannot empty')


def validate_base_url(base_url: str):
    if type(base_url) is not str:
        raise InvalidBaseURL('base_url can only be of type String')
    if base_url == '':
        raise InvalidBaseURL('base_url cannot empty')


def validate_storage_url(storage_url: str):
    if type(storage_url) is not str:
        raise InvalidBaseURL('storage_url can only be of type String')
    if storage_url == '':
        raise InvalidBaseURL('storage_url cannot empty')


def validate_deployment_name(deployment_name: str):
    if type(deployment_name) is not str:
        raise ValueError('deployment_name can only be of type String')
    if deployment_name == '':
        raise ValueError('deployment_name cannot empty')


def validate_model_version(model_version: str):
    if type(model_version) is not str:
        raise ValueError('model_version can only be of type String')
    if model_version == '':
        raise ValueError('model_version cannot empty')


def validate_namespace_id(namespace_id: str):
    if type(namespace_id) is not str:
        raise ValueError('namespace_id can only be of type String')
    if namespace_id == '':
        raise ValueError('namespace_id cannot empty')


def validate_platform_id(platform_id: list[str]):
    if type(platform_id) is not list[str]:
        raise ValueError('platform_id can only be of type List')
    if platform_id == '':
        raise ValueError('platform_id cannot empty')


def validate_model_pipeline(model_pipeline: str):
    if type(model_pipeline) is not str:
        raise ValueError('image_name can only be of type String')
    if model_pipeline == '':
        raise ValueError('image_name cannot empty')


def validate_model_folder(model_folder_path: Path):
    if model_folder_path == Path(''):
        raise ValueError(f"Model folder path is empty")

    if not model_folder_path.exists():
        raise FileNotFoundError(f"{model_folder_path} does not exist")

    if not model_folder_path.is_dir():
        raise NotADirectoryError(f"{model_folder_path} is not a folder")

    if not os.access(model_folder_path, os.W_OK):
        raise PermissionError(f"Folder doesn't have write permission: {model_folder_path}")


def validate_envs_args(name: str, val: list[str]):
    assert all([len(kv.split("=")) == 2 for kv in val]), f"{name} should be a string with comma-separated key value pairs. For e.g. 'k1=v1, k2=v2, k3=v3'"



class InvalidCredentialsError(Exception):
    def __init__(self, message):
        super().__init__(message)


class AuthenticationError(Exception):
    def __init__(self, message):
        super().__init__(message)


class ServerError(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidModelFolder(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidBaseURL(Exception):
    def __init__(self, message):
        super().__init__(message)


class InvalidStorageURL(Exception):
    def __init__(self, message):
        super().__init__(message)


class FileUploadError(Exception):
    def __init__(self, message):
        super().__init__(message)


class PipelineCreationError(Exception):
    def __init__(self, message):
        super().__init__(message)


class ModelDeploymentError(Exception):
    def __init__(self, message):
        super().__init__(message)


class BucketDoesNotExistsError(Exception):
    def __init__(self, message):
        super().__init__(message)



from typing import Union
from pathlib import Path
import minio
import os
from datetime import timedelta
from .errors import FileUploadError, InvalidCredentialsError, BucketDoesNotExistsError, PipelineCreationError
from zipfile import ZipFile
from .utils import api_error, Progress, get_all_file_paths, validate_url, validate_platform_id, \
    validate_namespace_id, validate_secret_key, validate_envs_args, \
    validate_model_folder
import requests
import json
import time

STORAGE_URL = "datalake.zbl-aws.zeblok.com:9000"
CROSS_CLOUD_SERVICE_URL = "http://cross-cloud-services-673735588.us-east-2.elb.amazonaws.com/upload"


class Pipeline:
    __slots__ = [
        '__base_url', '__cross_cloud_service_url', '__storage_url', '__api_access_key', '__api_access_secret',
        '__storage_username', '__storage_access_secret'
    ]

    def __init__(
            self, base_url: str, cross_cloud_service_url: str, storage_url: str, api_access_key: str,
            api_access_secret: str, storage_username: str, storage_access_secret: str
    ):

        validate_url(url_name='base_url', url=base_url)
        validate_url(url_name='cross_cloud_service_url', url=cross_cloud_service_url)
        validate_url(url_name='storage_url', url=storage_url)

        validate_secret_key(key_name='api_access_key', secret_key=api_access_key)
        validate_secret_key(key_name='api_access_secret', secret_key=api_access_secret)
        validate_secret_key(key_name='storage_username', secret_key=storage_username)
        validate_secret_key(key_name='storage_access_secret', secret_key=storage_access_secret)

        self.__base_url = 'https://' + base_url
        self.__cross_cloud_service_url = 'https://' + cross_cloud_service_url
        self.__storage_url = storage_url

        self.__api_access_key = api_access_key
        self.__api_access_secret = api_access_secret
        self.__storage_username = storage_username
        self.__storage_access_secret = storage_access_secret

    @property
    def base_url(self):
        return self.__base_url

    @property
    def cross_cloud_service_url(self):
        return self.__cross_cloud_service_url

    def get_datacenters(self) -> None:
        response = requests.get(
            f"{self.__base_url}/api/v1/dataCenter/public/", data=json.dumps({"isActive": True}),
            auth=(self.__api_access_key, self.__api_access_secret),
            headers={'Content-Type': 'application/json'}
        )
        if response.status_code == 200 and response.json()['success']:
            for data in response.json()['data']:
                print(f"Datacenter Name: {data['name']} | Datacenter ID: {data['_id']}")
        else:
            api_error(status_code=response.status_code, message=response.text)

    def get_namespaces(self) -> [dict]:
        response = requests.get(
            f"{self.__base_url}/api/v1/namespaces/org/", data=json.dumps({"state": "ready"}),
            auth=(self.__api_access_key, self.__api_access_secret),
            headers={'Content-Type': 'application/json'}
        )
        if response.status_code == 200 and response.json()['success']:
            for data in response.json()['data']:
                print(f"Namespace Name: {data['name']} | Namespace ID: {data['_id']}")
        else:
            api_error(status_code=response.status_code, message=response.text)

    def get_plans(self) -> None:
        response = requests.get(
            f"{self.base_url}/api/v1/plans/",
            auth=(self.__api_access_key, self.__api_access_secret),
            headers={'Content-Type': 'application/json'}
        )
        if response.status_code == 200 and response.json()['success']:
            for plan in response.json()['data']:
                print(f"Plan Name: {plan['planName']} | Id: {plan['_id']}")
                print("\tResources")
                print(f"\t\tvCPU: {plan['resources']['CPU']}")
                print(f"\t\tGPU: {plan['resources']['GPU']}")
                print(f"\t\tmemory [GB]: {plan['resources']['memory']}")
                print(f"\t\tstorage [GB]: {plan['resources']['storage']}")
                print(f"\t\tworkers: {plan['resources']['workers']}")
                print(f"\tDatacenter Name: {plan['dataCenterId']['name']} | Datacenter ID: {plan['dataCenterId']['_id']}")

        else:
            api_error(status_code=response.status_code, message=response.text)

    def __get_dockerfile(self, envs: list[str], entrypoint: list[str]) -> str:

        response = requests.get(
            f"{self.__base_url}/api/v1/system/dockerfile", auth=(self.__api_access_key, self.__api_access_secret),
            headers={'Content-Type': 'application/json'}, data=json.dumps({
                "from": {"baseImage": "miniconda3"}, "copy": {"./": "./"},
                "entrypoint": entrypoint, "env": dict([kv.split("=") for kv in envs])
            })
        )
        if response.status_code == 200:
            if response.json()['success']:
                return response.json()['data']
            else:
                api_error(status_code=response.status_code, message=response.json()['error']['message'])
        else:
            api_error(status_code=response.status_code, message=response.text)

    def __put_dockerfile(self, model_folder_path: Path, envs: list[str], entrypoint: list[str]) -> bool:
        validate_model_folder(model_folder_path=model_folder_path)
        with open(model_folder_path.joinpath('Dockerfile'), 'w') as fp:
            fp.write(self.__get_dockerfile(envs=envs, entrypoint=entrypoint))
        return True

    @staticmethod
    def __prepare_model_zip(model_folder_path: Path) -> Path:
        print("\nPreparing model zip .....")
        model_zipfile_path = model_folder_path.parent.joinpath(f'{model_folder_path.name.lower()}.zip')
        with ZipFile(model_zipfile_path, 'w') as zip:
            file_paths = get_all_file_paths(directory=model_folder_path)
            for file in file_paths:
                zip.write(filename=file.as_posix(), arcname=file.relative_to(model_folder_path))
        print("Model zip prepared")
        return model_zipfile_path

    def __upload_file(self, file_name: Path, bucket_name: str) -> str:
        try:
            client = minio.Minio(
                endpoint=self.__storage_url, access_key=self.__storage_username,
                secret_key=self.__storage_access_secret, secure=False
            )

            if not client.bucket_exists(bucket_name):
                raise BucketDoesNotExistsError(f'{bucket_name} does not exists')

            client.fput_object(
                bucket_name=bucket_name, object_name=f"{file_name.name}",
                file_path=file_name.as_posix(), content_type="application/zip",
                progress=Progress(), part_size=5 * 1024 * 1024
            )

            url = client.presigned_get_object(
                bucket_name=bucket_name, object_name=f"{file_name.name}",
                expires=timedelta(hours=3)
            )
        finally:
            os.remove(file_name)
        return url

    def __call_cross_cloud_service(
            self, cross_cloud_service_url: str, presigned_get_url: str, image_name: str, file_name: str,
            image_id: Union[str, None], plan_id: list[str], deployment_type: str, autodeploy: bool = False,
            namespace_id: Union[None, str] = None, platform_id: Union[None, list[str]] = None
    ):
        if autodeploy:
            # validate_platform_id(platform_id=platform_id)
            validate_namespace_id(namespace_id=namespace_id)

        response = requests.post(
            cross_cloud_service_url, auth=(self.__api_access_key, self.__api_access_secret),
            headers={'Content-Type': 'application/json'},
            data=json.dumps({
                "imageName": image_name,
                "url": presigned_get_url,
                "imageId": image_id,
                "filename": file_name,
                "portalUrl": f"{self.__base_url}",
                "autoDeploy": autodeploy,
                "namespaceId": None if not autodeploy else namespace_id,
                "platform": None if not autodeploy else platform_id,
                "deploymentType": deployment_type,
                "planId": plan_id
            })
        )
        if response.status_code == 200:
            if response.json()['success']:
                return
            else:
                raise FileUploadError(response.json()['job'])
        else:
            api_error(status_code=response.status_code, message=response.text)

    def __create_pipeline(
            self, pipeline_name: str, plan_id: list[str], namespace_id: str, datacenter_id: list[str], envs: list[str],
            args: list[str], docker_image_tag: str
    ):
        response = requests.post(
            f"{self.base_url}/api/v1/pipeline/", auth=(self.__api_access_key, self.__api_access_secret),
            headers={'Content-Type': 'application/json'},
            data=json.dumps({
                "pipelineName": pipeline_name,
                "planId": plan_id,
                "namespaceId": namespace_id,
                "yamlString": "from",
                "platform": datacenter_id,
                "env": envs,
                "args": args,
                "dockerImageTag": docker_image_tag
            })
        )
        print(response.request.body)
        if response.status_code == 200:
            if response.json()['success']:
                return response.json()['data']['_id']
            else:
                raise PipelineCreationError(response.json())
        else:
            api_error(status_code=response.status_code, message=response.text)

    def deploy_pipeline(
            self, model_folder_path: str, bucket_name: str, pipeline_name: str, plan_id: list[str], namespace_id: str,
            datacenter_id: list[str], envs: Union[str, None] = None, args: Union[str, None] = None,
            entrypoint: Union[str, None] = None, autodeploy: bool = False
    ):

        model_folder_path = Path(model_folder_path)

        envs = [] if envs is None else list(map(str.strip, envs.split(',')))
        if len(envs):
            validate_envs_args(name='envs', val=envs)

        args = [] if args is None else list(map(str.strip, args.split(',')))
        if len(args):
            validate_envs_args(name='args', val=args)

        entrypoint = [] if entrypoint is None else list(map(str.strip, entrypoint.split()))

        # self.__put_dockerfile(model_folder_path=model_folder_path, envs=envs, entrypoint=entrypoint)

        docker_image_tag = f"zeblok/{pipeline_name.lower()}:{int(time.time())}"

        pipeline_id = self.__create_pipeline(
            pipeline_name=pipeline_name, plan_id=plan_id, namespace_id=namespace_id, datacenter_id=datacenter_id,
            envs=envs, args=args, docker_image_tag=docker_image_tag
        )

        print("Completed create_pipeline")

        model_zipfile_path = self.__prepare_model_zip(model_folder_path)
        print(model_zipfile_path)
        print("Completed model Zipping")

        presigned_url = self.__upload_file(file_name=model_zipfile_path, bucket_name=bucket_name)
        print("Completed zip file upload")

        self.__call_cross_cloud_service(
            cross_cloud_service_url=self.cross_cloud_service_url, presigned_get_url=presigned_url,
            image_name=docker_image_tag, file_name=model_zipfile_path.name, image_id=pipeline_id, autodeploy=autodeploy,
            namespace_id=namespace_id, platform_id=datacenter_id,
            deployment_type='pipeline', plan_id=plan_id
        )
        print("Completed cross-cloud service call")



from pip._internal.req import parse_requirements
from setuptools import setup, find_packages

VERSION = '0.0.7'
DESCRIPTION = 'Zeblok Python SDK'
LONG_DESCRIPTION = 'Zeblok Python SDK to interact the Zeblok Ai-MicroCloud'

setup(
    name="zeblok-py",
    version=VERSION,
    author="Karan Pathak",
    author_email="karan@dataturtles.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url="",
    packages=find_packages(),
    install_requires=[item.requirement for item in parse_requirements('requirements.txt', session=False)],
    python_requires=">=3.9",
    keywords=['python', 'zeblok'],
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)


from .deploy import DeployModel
from pathlib import Path
from .errors import InvalidModelFolder, FileUploadError, InvalidCredentialsError, ModelDeploymentError
from .utils import api_error, get_all_file_paths
import requests
import time
import json
from .utils import validate_model_folder


class ModelMicroService(DeployModel):
    def __init__(self, base_url: str, token: str, bucket_name: str, username: str, storage_url: str):
        super(ModelMicroService, self).__init__(
            base_url=base_url, token=token, bucket_name=bucket_name, username=username, storage_url=storage_url
        )

    def __put_dockerfile(self, model_folder_path: Path) -> bool:
        validate_model_folder(model_folder_path=model_folder_path)
        with open(model_folder_path.joinpath('Dockerfile'), 'w') as fp:
            fp.write(self.__get_dockerfile(self.base_url, self.token))
        return True

    @staticmethod
    def __get_dockerfile(base_url: str, token: str) -> str:
        response = requests.get(
            f"{base_url}/api/v1/system/dockerfile",
            headers={'Authorization': token, 'Content-Type': 'application/json', 'request-from': 'Ai-MicroCloud'}
        )
        if response.status_code == 200:
            if response.json()['success']:
                return response.json()['data']
            else:
                api_error(status_code=response.status_code, message=response.json()['error']['message'])
        else:
            api_error(status_code=response.status_code, message=response.text)

    def get_plans(self) -> None:
        response = requests.get(
            f"{self.base_url}/api/v1/plans/",
            headers={'Authorization': self.token, 'Content-Type': 'application/json', 'request-from': 'Ai-MicroCloud'}
        )
        if response.status_code == 200 and response.json()['success']:
            for plan in response.json()['data']:
                print(f"Plan Name: {plan['planName']}")
                print(f"Id: {plan['_id']}")
                print(f"Datacenters:")
                print(f"\tName: {plan['dataCenterId']['name']}")
                print(f"\tId: {plan['dataCenterId']['_id']}")
        else:
            api_error(status_code=response.status_code, message=response.text)

    def __create_microservice(
            self, microservice_name: str, microservice_description: str, image_name: str, plan_id: str
    ) -> str:
        response = requests.post(
            f"{self.base_url}/api/v1/microservices/",
            data=json.dumps({
                "name": microservice_name,
                "description": microservice_description,
                "imageTag": [{"displayName": image_name, "dockerImage": image_name}],
                "defaultPlan": plan_id,
                "plans": [plan_id],
                "parameters": {
                    "ports": "4200"
                },
                "s3ImageLink": "https://zeblokcomputationalpublicimages.s3.ap-south-1.amazonaws.com/logo_6238a5205888964d61cdd42a.jpg",
                "isPublic": "true"
            }),
            headers={'Authorization': self.token, 'Content-Type': 'application/json', 'request-from': 'Ai-MicroCloud'}
        )
        # print(response.request.body)
        if response.status_code == 201:
            if response.json()['success']:
                print([{"displayName": image_name, "dockerImage": image_name}])
                return response.json()['data']['_id']
            else:
                raise ModelDeploymentError(response.json()['message'])
        else:
            api_error(status_code=response.status_code, message=response.text)

    def upload_model(
            self, model_folder_path: str, storage_secret_key: str, microservice_name: str,
            microservice_description: str, plan_id: str
    ) -> str:
        folder_path = Path(model_folder_path)
        # self.__put_dockerfile(Path(folder_path))
        model_zipfile_path = self.prepare_model_zip(model_folder_path=folder_path)
        presigned_url = self.upload_file(file_name=model_zipfile_path, secret_key=storage_secret_key)
        version = int(time.time())
        image_name = f"zeblok/{model_zipfile_path.stem}:{version}".lower()
        self.call_cross_cloud_service(
            cross_cloud_service_url=self.cross_cloud_service_url, presigned_get_url=presigned_url,
            image_name=image_name, file_name=model_zipfile_path.name, image_id=None, autodeploy=False,
            namespace_id='', deployment_name='', platform_id='', deployment_type='microservices'
        )
        return self.__create_microservice(
            microservice_name=f"{microservice_name}: {version}", microservice_description=microservice_description,
            image_name=image_name, plan_id=plan_id
        )

    def deploy_microservice(
            self, image_name: str, microservice_id: str, plan_id: str, service_name: str,
            namespace_id: str, datacenter_id: str, envs: list[str]
    ):

        response = requests.post(
            f"{self.base_url}/api/v1/spawnedservices/",
            data=json.dumps({
                "microserviceId": microservice_id, "dataCenterId": datacenter_id,
                "planId": plan_id,
                "namespaceId": namespace_id,
                "parameters": {
                    "ports": [5000],
                    "envs": envs,
                    "args": None,
                    "command": None
                },
                "name": service_name,
                "imageTag": {"displayName": image_name, "dockerImage": image_name}
            }),
            headers={'Authorization': self.token, 'Content-Type': 'application/json', 'request-from': 'Ai-MicroCloud'}
        )

        if response.status_code == 201:
            if response.json()['success']:
                return response.json()['message']
            else:
                raise ModelDeploymentError(response.json()['message'])
        else:
            api_error(status_code=response.status_code, message=response.text)
